# coding=utf-8
import sys, os, re
import subprocess
import time
from glob import glob
import pyliftover
from pysam import faidx, bcftools, FastxFile, AlignedSegment
from typing import Tuple, List


def simulate_reads_pbsim2(
        reference_sequence: str, path_to_pbsim: str = 'pbsim', depth: int = 1,
        simulation_mode: str = 'pbsim2/data/R103.model',
        difference_ratio: str = '23:31:46', length_mean: int = 20000,
        length_max: int = 1000000, length_min: int = 100,
        length_sd: int = 15000, accuracy_min: float = 0.01,
        accuracy_max: float = 1.00, accuracy_mean: float = 0.80,
        prefix: str = None, id_prefix: str = 'S', output_directory: str = None, sample: str = None,
        haplotype: int = None, quiet = False, silent = False
        ) -> str:
    """
    Args:
        reference_sequence (str):
            reference sequence template to simulate reads from
        path_to_pbsim (str):
            if pbsim2 is not installed to the PATH as pbsim then use this to set its absolute or relative path
        depth (int):
            the depth of coverage that will be simulated
        simulation_mode: 
        difference_ratio: 
        length_mean: 
        length_min: 
        length_max: 
        length_sd: 
        accuracy_min: 
        accuracy_max: 
        accuracy_mean: 
        prefix: 
        id_prefix: 
        output_directory: 
        sample: 
        haplotype: 

    Returns:
        object: 
    
    """
    
    if output_directory is None:
        output_directory: str = 'simulated'
    
    if not os.path.exists(output_directory):
        os.mkdir(output_directory)
    
    if not sample is None and not haplotype is None:
        if prefix is None:
            prefix = str(output_directory) + '/' + str(sample) + '__' + str(haplotype) + '__' + 'pbsim2_simulated_' +\
                     os.path.basename(reference_sequence).split('.')[0] + '_' + str(time.localtime()[0]) + '_' + str(
                    time.localtime()[1]
                    ) + '_' + str(time.localtime()[2]) + '_' + str(time.localtime()[3]) + 'hr_' + str(
                    time.localtime()[4]
                    ) + 'min_' + str(time.localtime()[5]) + 'sec'
        else:
            prefix = '{0}/{1}__{2}__{3}'.format(str(output_directory), str(sample), str(haplotype), str(prefix))
    else:
        prefix = '{0}/{1}__{2}__{3}'.format(str(output_directory), str(sample), str(haplotype), str(prefix))
    
    if str(simulation_mode).lower().endswith(".fastq"):
        mode1 = '--sample-fastq'
        mode2 = simulation_mode
    
    elif str(simulation_mode).lower().endswith('.model'):
        mode1 = '--hmm_model'
        mode2 = simulation_mode
    
    if not reference_sequence:
        sys.stderr.write('No reference genome was provided for pbsim2 simulation. Please specify reference_genome_path.\n')
        raise exception('Reference sequence not provided.')
        return
    
    sys.stderr.write('\n################ Beginning Read Simulation ################\n')
    
    start_process_time = time.time()
    
    reference_genome_path = os.path.abspath(reference_sequence)
    reference_genome_index_path = str(reference_genome_path) + '.fai'
    if not os.path.isfile(reference_genome_index_path):
        faidx(reference_genome_path)

    pbsim2_args = [str(path_to_pbsim), f'--depth', str(depth),
                   str(mode1), str(mode2),
                   f'--difference-ratio', str(difference_ratio),
                   f'--length-mean', str(length_mean),
                   f'--length-min', str(length_min),
                   f'--length-max', str(length_max),
                   f'--length-sd', str(length_sd),
                   f'--accuracy-mean', str(accuracy_mean),
                   f'--accuracy-min', str(accuracy_min),
                   '--accuracy-max', str(accuracy_max),
                   f'--prefix', str(prefix),
                   '--id-prefix', str(id_prefix),
                   str(reference_genome_path)]

    if silent:
        subprocess.run(pbsim2_args, check = True, stdout = subprocess.DEVNULL, stderr = subprocess.DEVNULL)
    elif quiet:
        subprocess.run(pbsim2_args, check = True, stdout = subprocess.DEVNULL)
    else:
        subprocess.run(pbsim2_args, check = True)

    sys.stderr.write('Post-processing simulated fastq files...\n')
    sample_hap_to_true_alignment_dict = build_true_alignment_dict(reference_genome_index_path, prefix)

    final_combined_simulated_long_reads_fastq_path = '{0}/{1}.fastq'.format(
        str(output_directory),
        str(os.path.splitext(os.path.basename(reference_sequence))[0])
    )    
    for fastq_file in glob(str(prefix) + '*.fastq'):
        if not (quiet or silent):
            sys.stderr.write("%s\n" % fastq_file)
        with FastxFile(fastq_file, 'r') as infile:
            with open(final_combined_simulated_long_reads_fastq_path, 'a') as outfile:
                for read in infile:
                    read.name = str(sample) + '__' + str(haplotype) + '__' + str(sample_hap_to_true_alignment_dict[read.name])
                    outfile.write(str(read) + '\n')
        os.remove(fastq_file)

    # Clean up pbsim2 output files
    for f in glob(str(prefix) + '*.maf'):
        os.remove(f)
    for f in glob(str(prefix) + '*.ref'):
        os.remove(f)

    sys.stderr.write('Simulation finished in %.2f seconds.\n' % (time.time() - start_process_time))
    return final_combined_simulated_long_reads_fastq_path


def build_true_alignment_dict(reference_genome_index_path, prefix):
    """
    Build the true alignment dict from pbsim2 maf files.
    """
    true_alignment_dict = {}
    files = glob(prefix + '*.maf')
    for pbsim2_maf in files:
        al_dict = extract_true_alignments_from_pbsim2_maf(reference_genome_index_path, pbsim2_maf)
        true_alignment_dict.update(al_dict)
    return true_alignment_dict


def extract_true_alignments_from_pbsim2_maf(reference_genome_index_path, pbsim2_maf):
    """
    Parse true alignment coordinates from a pbsim2 MAF file.
    """
    # Build the chromosome list from the reference genome index
    chr_list = []
    with open(reference_genome_index_path, 'r') as index_file:
        for rec in index_file:
            line = rec.strip('\n').split()
            if len(line) > 0:
                chr_list.append(line[0])

    # Parse MAF records, extracting the true alignment coordinates for each simulated read.
    true_alignment_dict = {}
    state = 0
    reg = []
    with open(pbsim2_maf) as maf_file:
        for rec in maf_file:
            line = rec.strip('\n').split()
            if len(line) == 0:
                continue
            elif state == 0 and line[0] == 'a':
                state = 1
            elif state == 1 and line[0] == 's':
                reg = [line[2], line[2]+line[3]]
                state = 2
            elif state == 2 and line[0] == 's':
                m = re.match('^S\d+_\d+', line[1])
                if not m:
                    sys.stderr.write("Failed to parse read name from record: %s. Skipping...\n" % " ".join(line[0:5]))
                    state = 0
                    continue
                chr_id = int(line[1].split('S')[1].split('_')[0]) - 1
                if chr_id > len(chr_list):
                    sys.stderr.write("Chromosome not in index: %s. Skipping...\n" % str(chr_id+1))
                    state = 0
                    continue
                name = "!".join([line[1], chr_list[chr_id], reg[0], reg[1], line[4]])
                seq = line[6].replace('-', '')
                if len(seq) != int(line[5]):
                    sys.stderr.write("Inconsistent read length for sequence %s. Skipping...\n" % name)
                    state = 0
                    continue
                true_alignment_dict[line[1]] = name
                #sys.stderr.write("%s, %s\n" % (line[1], name))
                state = 0

    return true_alignment_dict
    

def convert_coordinates(
        haplotype: int, sample: str, contig: str, position: int, reference_sequence: str,
        reference_liftover_converters = None, chain_file = None
        ):
    """

    Args:
        haplotype (int):
        sample (str):
        contig (str):
        position (int):
        reference_sequence (str):
        reference_liftover_converters: 
        chain_file: 

    Returns:
        object: 

    """
    if reference_liftover_converters is None:
        reference_liftover_converters = {}
    
    if reference_sequence in reference_liftover_converters:
        if sample in reference_liftover_converters[reference_sequence]:
            if str(haplotype) in reference_liftover_converters[reference_sequence][sample]:
                return reference_liftover_converters[reference_sequence][sample][str(haplotype)].convert_coordinates(
                        contig, position
                        )
        
        elif chain_file is not None:
            if os.path.isfile(chain_file):
                reference_liftover_converters[reference_sequence][sample] = {}
                reference_liftover_converters[reference_sequence][sample][str(haplotype)] = pyliftover.LiftOver(
                        chain_file
                        )
                return reference_liftover_converters[reference_sequence][sample][str(haplotype)].convert_coordinates(
                        contig, position
                        )
    
    elif chain_file is not None:
        if os.path.isfile(chain_file):
            reference_liftover_converters[reference_sequence] = {}
            reference_liftover_converters[reference_sequence][sample] = {}
            reference_liftover_converters[reference_sequence][sample][str(haplotype)] = pyliftover.LiftOver(chain_file)
            return reference_liftover_converters[reference_sequence][sample][str(haplotype)].convert_coordinates(
                    contig, position
                    )


def generate_haplotype_specific_fasta(
        haplotype: int, sample: str, input_reference_sequence_path: str, haplotype_vcf: str,
        output_reference_sequence_path: str = None,
        chain_file_path: str = None
        ) -> object:
    """

    Args:
        haplotype:
        sample:
        input_reference_sequence_path:
        haplotype_vcf:
        output_reference_sequence_path:
        chain_file_path:

    Returns:

    """
    sys.stderr.write('\n################ Generating Reference Sequence for Haplotype %d ################\n' % haplotype)
    input_reference_sequence_path = os.path.abspath(input_reference_sequence_path)
    input_reference_sequence_index_path: str = str(input_reference_sequence_path) + '.fai'

    if output_reference_sequence_path is None:
        output_reference_sequence_path: str = str(os.path.splitext(input_reference_sequence_path)[0]) + '_hap' + str(
                haplotype
                ) + '_' + str(sample) + '.fa'
    
    if chain_file_path is None:
        chain_file_path = str(os.path.splitext(input_reference_sequence_path)[0]) + '_hap' + str(haplotype) + '_' + str(
                sample
                ) + '.chain'

    # This causes python to segfault if bcftools encounters an error!
    fasta_output = bcftools.consensus(
            '-H', str(haplotype), '-s', str(sample), '-c', str(chain_file_path), '-f',
            str(input_reference_sequence_path),
            str(haplotype_vcf)
            )
    with open(output_reference_sequence_path, 'w') as output_fasta:
        output_fasta.write(fasta_output)

    sys.stderr.write("Done.\n")
    return output_reference_sequence_path, chain_file_path


def index_fasta_file(fasta_file_path: str) -> str:
    """

    Args:
        fasta_file_path: 

    Returns:

    """
    fasta_file_path = os.path.abspath(fasta_file_path)
    
    if fasta_file_path.endswith('.fa') or fasta_file_path.endswith('fa.gz'):
        faidx(fasta_file_path)
        return os.path.abspath(fasta_file_path)
    
    elif fasta_file_path.endswith('fa.gz'):
        os.rename(fasta_file_path, '{0}.fa.gz'.format(str(os.splitext(fasta_file_path)[0])))
        faidx(str(os.splitext(fasta_file_path)[0]) + '.fa.gz')
        return str(os.splitext(fasta_file_path)[0]) + '.fa.gz'
    
    else:
        os.rename(fasta_file_path, str(os.splitext(fasta_file_path)[0]) + '.fa')
        faidx(str(os.splitext(fasta_file_path)[0]) + '.fa')
        return str(os.splitext(fasta_file_path)[0]) + '.fa'


def create_fasta_file(
        input_fasta_file_path: str, output_fasta_file_path: str = None, regions: List[str] = None, only_autosomal:
        bool = False, chrom_sizes: str = None
        ) -> str:
    """

    Args:
        input_fasta_file_path: 
        only_autosomal (object): 
        regions:
            List of regions and/or contigs ie: ['chr1:10000000:60000000',...]
        output_fasta_file_path (object): 
    """
    regions_file: str = str(os.path.splitext(output_fasta_file_path)[0]) + '_regions.txt'
    autosomal_reference_names = ['chr1', 'chr2', 'chr3', 'chr4', 'chr5', 'chr6', 'chr7', 'chr8', 'chr9', 'chr10',
                                 'chr11', 'chr12', 'chr13', 'chr14', 'chr15', 'chr16', 'chr17', 'chr18', 'chr19',
                                 'chr20', 'chr21', 'chr22'] # This should be generalized for non-human genomes
    input_fasta_file_path = os.path.abspath(input_fasta_file_path)
    
    if output_fasta_file_path is None:
        if only_autosomal:
            output_fasta_file_path = str(
                    os.path.splitext(input_fasta_file_path)[0]
                    ) + 'output_regions_only_autosomal.fa'
        else:
            output_fasta_file_path = str(os.path.splitext(input_fasta_file_path)[0]) + 'output_regions.fa'
    
    if regions is None and not only_autosomal:
        """
        If we are not filtering out any specific regions, just test for a fai file
        and create one if not found. Return the output_fasta_file_path.
        """
        if not os.path.isfile(input_fasta_file_path + '.fai'):
            # This always creates the index in the directory where the fasta file lives.
            # Don't create a copy of the reference if we're not filtering anything!
            faidx(input_fasta_file_path, '-o', output_fasta_file_path)
        return input_fasta_file_path

    # Otherwise, we need to filter out sequences we want and reindex the resulting fasta.
    if regions is None and only_autosomal:
        # Explicitly providing regions will always supercede only_autosomal
        if chrom_sizes == None:
            # We need chrom.sizes because the regions file must have start and end coordinates. See http://www.htslib.org/doc/samtools-faidx.html
            raise Exception("ERROR: create_fasta_file only_autosomal requires chrom_sizes.")
            return
        with open(regions_file, 'w') as rg_file:
            with open(chrom_sizes, 'r') as cs_file:
                for line in cs_file:
                    rec = line.split()
                    if rec[0] in autosomal_reference_names:
                        # Write a line to the regions file.
                        rg_file.write("{}:1-{}\n".format(rec[0], rec[1]))

    elif os.path.isfile(regions):
        regions_file = regions
        
    elif isinstance(regions, list):
        if only_autosomal:
            regions_list = []
            for region in regions:
                chrom = region.split(':')[0]
                if chrom in autosomal_reference_names:
                    regions_list.append(region)                
        else:
          regions_list = regions
            
        # Validate all entries in regions_list
        chrom_sizes_dict = {}
        with open(chrom_sizes, 'r') as cs_file:
            for line in cs_file:
                rec = line.split()
                chrom_sizes_dict[rec[0]] = rec[1]

        with open(regions_file, 'w') as file:
            for region in regions_list:
                # Quick and dirty validation -- just checks for presence of the proper delimiters
                # and end coordinate within the extent of the chromosome.
                s1 = region.split(':')
                s2 = region.split('-')
                if len(s1) == 2 and len(s2) == 2 and s1[0] in chrom_sizes_dict.keys() and s2[1] <= chrom_sizes_dict[s1[0]]:
                    file.write(str(region) + '\n')
                else:
                    if s1[0] in chrom_sizes_dict.keys() and s2[1] <= chrom_sizes_dict[s1[0]]:
                        sys.stderr.write('WARNING: Region {} coordinates not found in the reference sequence. Omitting.\n'.format(region))
                    else:
                        sys.stderr.write('WARNING: Region {} is not in the proper format (chrom:start-end). Omitting.\n'.format(region))

    faidx(input_fasta_file_path, '-r', regions_file, '-o', output_fasta_file_path)
    
    return output_fasta_file_path


def true_alignment_match(
        aligned_segment: AlignedSegment, needs_liftover: bool = False, contig: str = None, ref_start: int = None,
        ref_end: int = None, strand: str = None, tag_read: bool = False, only_output_match_label: bool = False, only_output_overlap: bool = False,
        true_reference_sequence_path: str = None, aligned_reference_sequence_path: str = None, liftover_converter = None,
        chain_file_path: str = None, sample: str = None, haplotype: int = None, reference_liftover_converters = None
        ) -> object:
    """

    Args:
        chain_file_path (object): 
        aligned_segment: 
        sample: 
        tag_read (object): 
    """
    overlap = 0
    match_label = 'non_match'
    
    if contig is None or ref_start is None or ref_end is None or strand is None:
        contig, ref_start, ref_end, strand = true_alignment_coordinates(
                aligned_segment, tag_read = tag_read, contig = contig, ref_start = ref_start, ref_end = ref_end,
                strand = strand
                )
    
    if needs_liftover:
        if liftover_converter is not None:
            liftover_converter = liftover_converter
    
        elif reference_liftover_converters is not None:
            if aligned_reference_sequence_path is not None:
                if aligned_reference_sequence_path in reference_liftover_converters:
                    if sample in reference_liftover_converters[aligned_reference_sequence_path]:
                        if str(haplotype) in reference_liftover_converters[aligned_reference_sequence_path][sample]:
                            liftover_converter = reference_liftover_converters[aligned_reference_sequence_path][sample][
                                str(haplotype)]
    
        elif chain_file_path is not None:
            if os.path.isfile(chain_file_path):
                liftover_converter = pyliftover.Liftover(chain_file_path)
    
        else:
            sys.stderr.write('Could not liftover\n')
            return
    
        contig = liftover_converter.convert_coordinate(contig, ref_start)[0][0]
        ref_start = liftover_converter.convert_coordinate(contig, ref_start)[0][1]
        ref_end = liftover_converter.convert_coordinate(contig, ref_end)[0][1]
        strand = liftover_converter.convert_coordinate(contig, ref_start)[0][2]
    
    if aligned_segment.reference_name == contig:
        match_label = 'ref_match'
        if int(ref_start) <= int(aligned_segment.reference_start) <= int(ref_end):
            match_label = 'mapping_match'
            if int(aligned_segment.reference_end) >= int(ref_end):
                overlap = int(ref_end) - int(aligned_segment.reference_start)
            else:
                overlap = int(aligned_segment.reference_end) - int(aligned_segment.reference_start)
        elif int(ref_start) <= int(aligned_segment.reference_end) <= int(ref_end):
            match_label = 'mapping_match'
            if int(ref_start) >= int(aligned_segment.reference_start):
                overlap = int(aligned_segment.reference_end) - int(ref_start)
            else:
                overlap = int(aligned_segment.reference_end) - int(aligned_segment.reference_start)
        elif (
                int(ref_start) - (int(ref_start) * 0.1)) <= aligned_segment.reference_start <= (
                int(ref_end) + (int(ref_end) * 0.1)):
            match_label = 'within_10percent'
    
    if tag_read:
        aligned_segment.set_tag(
                tag = 'ov', value = str(overlap / (int(ref_end) - int(ref_start))), value_type = 'Z', replace = True
                )
        aligned_segment.set_tag(
                tag = 'OV', value = str(overlap / (aligned_segment.query_alignment_length)), value_type = 'Z',
                replace = True
                )
        aligned_segment.set_tag(tag = 'ml', value = str(match_label), value_type = 'Z', replace = True)
    
    if only_output_match_label:
        return match_label
    
    elif only_output_overlap:
        return overlap
    
    else:
        return match_label, overlap


def true_alignment_coordinates(
        aligned_segment: AlignedSegment, tag_read: bool = False, contig: str = None, ref_start: int = None,
        ref_end: int = None, strand: str = None
        ) -> object:
    """

    Args:
        aligned_segment: 
        tag_read: 
        contig: 
        ref_start: 
        ref_end: 
        strand: 

    Returns:

    """
    if '!' in str(aligned_segment.query_name):
        _read_name, _contig, _ref_start, _ref_end, _strand = aligned_segment.query_name.split('!')
        if contig is None:
            contig = _contig
        if ref_start is None:
            ref_start = _ref_start
        if ref_end is None:
            ref_end = _ref_end
        if strand is None:
            strand = _strand
    
    if tag_read:
        aligned_segment.set_tag(tag = 'st', value = str(ref_start), value_type = 'Z', replace = True)
        aligned_segment.set_tag(tag = 'lo', value = str(contig), value_type = 'Z', replace = True)
        aligned_segment.set_tag(tag = 'en', value = str(ref_end), value_type = 'Z', replace = True)
        aligned_segment.set_tag(tag = 'sd', value = str(strand), value_type = 'Z', replace = True)
    
    return contig, ref_start, ref_end, strand


def alignment_type(aligned_segment: AlignedSegment, tag_read: bool = False) -> str:
    """

    Args:
        aligned_segment:
        tag_read:

    Returns:

    """
    alignment_label: str = 'None'
    if aligned_segment.is_unmapped:
        alignment_label = 'unmapped'
    elif aligned_segment.is_secondary:
        alignment_label = 'secondary'
    elif aligned_segment.is_supplementary:
        alignment_label = 'supplementary'
    else:
        alignment_label = 'mapped'
    if tag_read:
        aligned_segment.set_tag(tag = 'al', value = str(alignment_label), value_type = 'Z', replace = True)
    
    return alignment_label



from collections import defaultdict
from decimal import Decimal
from fractions import Fraction
from statistics import mean
from typing import Any, List, Optional, Tuple, Union

from Bio import Align


# noinspection PyUnusedClass
class HeterozygousSite:
    #probs_read: List[Union[float, Any]]
    #aligned_het_bases: defaultdict[Any, dict]
    #total_non_matches: int
    #aligned_het_bases: defaultdict[Any, dict]
    #local_realignment_counter: defaultdict[Any, int]
    # best_haplotype: List[Any]
    #alignments_favored_score_for_vary_lengths: List[Tuple[Optional[Any], List[Any]]]
    #average_score_diff_for_local_alignments: Union[Union[float, Decimal, Fraction], Any]
    sample: str
    vcf_record: object
    
    def __init__(self, vcf_record, sample: object, aligned_segment, liftover_converters = None):
        """
        Args:
            vcf_record:
            sample (object):
            aligned_segment:
            liftover_converters:
        """
        self.total_non_matches = None
        self.probs_read = None
        self.haplotype = None
        self.vcf_record = vcf_record
        self.sample = sample
        self.aligned_segment = aligned_segment
        if liftover_converters:
            self.liftover_converters = liftover_converters
        else:
            self.liftover_converters = None
    
    @property
    def is_snv(self) -> bool:
        """
        Returns:
            bool:
        """
        SNV: bool = True
        for haplotype in range(0, self.ploidy_het_site):
            if not len(self.alleles[haplotype]) == 1:
                SNV = False
        return SNV
    
    @property
    def is_match(self) -> bool:
        match = False
        for haplotype in range(0, self.ploidy_het_site):
            if self.alleles[haplotype] == self.read_base:
                match = True
        
        return match
    
    @property
    def is_match_to_phase(self) -> List[bool]:
        """
        Returns:
            object:
        """
        matches: List[bool] = []
        for haplotype in range(0, self.ploidy_het_site):
            match: bool = False
            if self.alleles[haplotype] == self.read_base:
                match = True
            matches.append(match)
        
        return matches
    
    @property
    def matches_single_phase(self) -> bool:
        """
        Returns:
            bool:
        """
        matches = []
        for haplotype in range(0, self.ploidy_het_site):
            match = False
            if self.alleles[haplotype] == self.read_base:
                match = True
            if match:
                matches.append(haplotype + 1)
        if len(matches) == 1:
            return True
        else:
            return False
    
    @property
    def haplotype_matches(self) -> object:
        """
        Returns:
            List[int]:
        """
        matches: List[int] = []
        haplotype: int
        for haplotype in range(0, self.ploidy_het_site):
            match = False
            if self.alleles[haplotype] == self.read_base:
                match = True
            if match:
                matches.append(haplotype + 1)
        return matches
    
    def is_non_match(self, haplotype: int = None) -> bool:
        """Args: int
            haplotype:

        Args:
            haplotype (int):

        Returns:
            bool:
        """
        
        # haplotype: int
        # for haplotype in range(0, self.ploidy_het_site):
        #     if not self.alleles[haplotype] == self.read_base:
        #         return False
        
        return self.alleles[haplotype] == self.read_base

    @property
    def ploidy_het_site(self) -> int:
    
        """INPUT ATTRIBUTE = FALSE OUTPUT ATTRIBUTE = TRUE :returns: :rtype:
        int
        """
    
        return self.vcf_record.samples[str(self.sample)].alleles
    
        # assert isinstance(self.alleles, object)
        # return len(self.alleles)
    
    @property
    def is_heterozygous(self) -> bool:
        """
        Returns:
            bool:
        """
        return min(self.vcf_record.samples[str(self.sample)].allele_indices) < max(
                self.vcf_record.samples[str(self.sample)].allele_indices
                )
    
    @property
    def reference_position(self) -> int:
        """
        Returns:
            int:
        """
        return self.vcf_record.pos - 1
    
    @property
    def reference_name(self) -> str:
        """
        Returns:
            str:
        """
        return str(self.aligned_segment.reference_name)
    
    @property
    def alleles(self) -> List[str]:
        """
        Returns:
            List[str]:
        """
        return self.vcf_record.samples[str(self.sample)].alleles
    
    @property
    def is_phased(self) -> bool:
        """
        Returns:
            bool:
        """
        return self.vcf_record.samples[str(self.sample)].phased
    
    @property
    def gapped_alignment_position(self) -> int:
        """
        Returns:
            int:
        """
        
        self._generate_gapped_alignment_position()
        
        return self._gapped_alignment_position
    
    @property
    def read_base(self) -> str:
        """
        Returns:
            str:
        """
        return self.aligned_segment.query_sequence[self.gapped_alignment_position]
    
    @property
    def read_base_quality(self) -> bytes:
        """
        Returns:
            bytes:
        """
        return self.aligned_segment.query_qualities[self.gapped_alignment_position]
    
    @property
    def read_base_quality_error_rate(self) -> float:
        """
        Returns:
            float:
        """
        return 10**(-0.1 * self.aligned_segment.query_qualities[self.gapped_alignment_position])
    
    # def read_base_quality_error_rate_for_offset(self, offset: object) -> object:
    #     """
    #
    #     Args:
    #         offset:
    #
    #     Returns:
    #
    #     """
    #     return 10**(-0.1 * self.aligned_segment.query_qualities[self.gapped_alignment_position + offset])
    
    # @property
    # def correct_phase(self) -> bool:
    #     """
    #
    #     SIMULATED SUBCLASS
    #     Returns:
    #         bool:
    #
    #     """
    #     if not self.aligned_segment.has_tag('oa'):
    #         return
    #     return self.aligned_segment.get_tag('oa')
    
    @property
    def max_allele_length(self) -> int:
        """
        Returns:
            int:
        """
        return max([len(allele) for allele in self.alleles])
    
    @property
    def allele_length(self) -> List[int]:
        return [len(allele) for allele in self.alleles]
    
    def allele_length_for_haplotype_i(self, haplotype) -> int:
        """
        Args:
            haplotype: int:

        Returns:
            int:
        """
        return [len(allele) for allele in self.alleles][int(haplotype) - 1]
    
    # @property
    def query_read_sequence(self, length: int) -> object:
        """
        Args:
            length (int):

        Returns:
            List[str]:
        """
        if self.gapped_alignment_position - length < 0:
            start = 0
        else:
            start = self.gapped_alignment_position - length
        
        if self.gapped_alignment_position + length - self.aligned_segment.query_length > 0:
            end = self.aligned_segment.query_length
        else:
            end = self.gapped_alignment_position + length
        
        return self.aligned_segment.query_sequence[start:end]
    
    # @property
    def target_ref_sequence(self, haplotype: object, length: object) -> object:
        """
        Args:
            haplotype (object): int
            length (object): int

        Returns:
            str:
        """
        if self.liftover_converters is None:
            return
        elif str(haplotype) in self.liftover_converters:
            # start = self.liftover_converters[str(haplotype)]['converter'].convert_coordinate(str(
            # self.reference_name), self.reference_position-length) print('start',start) if start: start = start[0][
            # 1] else: multiplier = 1 while len(start) == 0: start = self.liftover_converters[str(haplotype)][
            # 'converter'].convert_coordinate(str(self.reference_name), self.reference_position-(length*multiplier))
            # print('start',start, 'multiplier',multiplier) multiplier+=1 start = start[0][1] end =
            # self.liftover_converters[str(haplotype)]['converter'].convert_coordinate(str(self.reference_name),
            # self.reference_position+length) print('end',end) if end: end = end[0][1] else: multiplier = 1 while
            # len(end) == 0: end = self.liftover_converters[str(haplotype)]['converter'].convert_coordinate(str(
            # self.reference_name), self.reference_position+(length*multiplier)) print('end',end, 'multiplier',
            # multiplier) multiplier+=1 end = end[0][1] print('start',start) print('end',end)
            reference = str(self.reference_name)
            mid: object = self.liftover_converters[str(haplotype)]['converter'].convert_coordinate(
                    str(self.reference_name),
                    self.reference_position
                    )
            if len(mid) == 0:
                print(mid)
                multiplier = 1
                while len(mid) == 0:
                    mid = self.liftover_converters[str(haplotype)]['converter'].convert_coordinate(
                            str(self.reference_name), self.reference_position + multiplier
                            )
                    print('mid', mid, 'multiplier', multiplier)
                    multiplier += 1
            mid = mid[0][1]
            sequence: str = self.liftover_converters[str(haplotype)]['reference_fasta'].fetch(
                    reference = str(self.reference_name), start = mid - length, end = mid + length
                    )
            return sequence
    
    def local_realignment(self, length: object, upper_flag: object = True) -> object:
        """
        Args:
            length (object): int
            upper_flag (object): bool

        Returns:
            object:
        """
        aligner = Align.PairwiseAligner()
        aligner.mode = 'local'
        aligner.target_open_gap_score = -1
        aligner.query_open_gap_score = -1
        aligner.mismatch_score = -0.5
        aligner.extend_gap_score = -0.5
        haplotype_favored = None
        query = self.query_read_sequence(length)
        
        alignments = defaultdict(list)
        for haplotype in range(1, self.ploidy_het_site + 1):
            target = self.target_ref_sequence(str(haplotype), length)
            if upper_flag:
                alignments[str(haplotype)] = aligner.align(str(target).upper(), str(query).upper())
            else:
                alignments[str(haplotype)] = aligner.align(target, query)
        
        return alignments
    
    # @property
    @property
    def calculate_average_score_diff_for_local_alignments(self) -> object:
        """
        Returns:
            float:
        """
        average_score_diff_for_local_alignments: Union[Union[float, Decimal, Fraction], Any]
        
        length: int
        self.average_score_diff_for_local_alignments = mean(
                [pair[1][0] for pair in [
                        calculate_favored_haplotype_from_local_alignment(self.local_realignment(length)) for length
                        in
                        range(5, 100, 5)] if len(pair[1]) > 0]
                )
        return self.average_score_diff_for_local_alignments
    
    # @property
    @property
    def automated_evaluation_of_local_alignments(self) -> object:
        """
        Returns:
            object:
        """
        alignments_favored_score_for_vary_lengths: List[Tuple[Optional[Any], List[Any]]]
        
        self.alignments_favored_score_for_vary_lengths = []
        for length in range(5, 100, 5):
            self.alignments_favored_score_for_vary_lengths.append(
                    calculate_favored_haplotype_from_local_alignment(self.local_realignment(length))
                    )
        self.sorted_pos = sorted(
                [pos for pos in zip(self.alignments_favored_score_for_vary_lengths, list(range(5, 100, 5))) if
                 pos[0][0]],
                key = lambda x:x[1], reverse = False
                )
        return self.alignments_favored_score_for_vary_lengths
    
    # @property
    @property
    def select_best_haplotype(self) -> int:
        """
        Returns:
            int:
        """
        best_haplotype: List[Any]
        self.automated_evaluation_of_local_alignments()
        self.count_best_haplotypes
        self.best_haplotype = []
        for haplotype in self.local_realignment_counter.items():
            haplotype: int
            if haplotype[1] == max(
                    [self.local_realignment_counter[haplotype] for haplotype in
                     self.local_realignment_counter]
                    ):
                self.best_haplotype.append(haplotype[0])
        if len(self.best_haplotype) == 1:
            return self.best_haplotype[0]
    
    # @property
    @property
    def count_best_haplotypes(self) -> int:
        """
        Returns:
            int:
        """
        local_realignment_counter: defaultdict[Any, int]
        self.local_realignment_counter = defaultdict(int)
        R: int
        for R in self.alignments_favored_score_for_vary_lengths:
            if 0 < len(R[1]):
                self.local_realignment_counter[R[0]] += R[1][0]
        if [self.local_realignment_counter[haplotype] for haplotype in self.local_realignment_counter if
            self.local_realignment_counter[haplotype]]:
            return max([self.local_realignment_counter[haplotype] for haplotype in self.local_realignment_counter])
    
    def calculate_aligned_het_bases_from_local_alignment(self, alignments: object) -> object:
        """
        Args:
            alignments (object): List[object]:

        Returns:
            int:
        """
        aligned_het_bases: defaultdict[Any, dict]
        self.aligned_het_bases: defaultdict[Any, dict] = defaultdict(dict)
        position: int
        for position in range(0, self.max_allele_length):
            query_position: int = self.gapped_alignment_position + position
            haplotype: object
            for haplotype in alignments:
                self.aligned_het_bases[str(haplotype)][query_position] = defaultdict(dict)
                
                # aligned_het_bases[haplotype][query_position]['target_base'] = defaultdict(list)
                # aligned_het_bases[haplotype][query_position]['query_base'] = defaultdict(list)
                
                index_length = int(len(alignments[str(haplotype)][0].target) / 2) + position
                if int(len(alignments[str(haplotype)][0].target) / 2) > position:
                    target_base = alignments[str(haplotype)][0].target[index_length]
                else:
                    target_base = '-'
                
                query_base: str = '-'
                for block in zip(alignments[str(haplotype)][0].aligned[0], alignments[str(haplotype)][0].aligned[1]):
                    if index_length in range(block[0][0], block[0][1]):
                        query_base = alignments[str(haplotype)][0].query[index_length - block[0][0] + block[1][0]]
                
                self.aligned_het_bases[haplotype][query_position][
                    'error_rate_quality_score'] = self.read_base_quality_error_rate_for_offset(position)
                self.aligned_het_bases[haplotype][query_position]['target_base'] = target_base
                self.aligned_het_bases[haplotype][query_position]['query_base'] = query_base
                self.aligned_het_bases[haplotype][query_position]['match'] = query_base == target_base
        
        return self.aligned_het_bases
    
    def calculate_probs(self, haplotype: object = None, force_recalc = False) -> float:
        """
        Args:
            haplotype (object):

        Returns:
            float:
        """
        # Use cached probs_read if possible.
        if len(self.probs_read) > 0 and not force_recalc:
            return sum(self.probs_read)
        
        total_non_matches: int = 0
        self.total_matches = 0
        #self.total_non_matches = 0
        aligned_het_bases: defaultdict[Any, dict]
        
        var = self.automated_evaluation_of_local_alignments
        if haplotype is None:
            haplotype = self.select_best_haplotype
        else:
            haplotype = haplotype
        
        var = self.automated_evaluation_of_local_alignments
        
        self.probs_read: List[Union[float, Any]] = []
        
        lengths: List[Any] = []
        for pos in self.sorted_pos:
            if pos[0][0] == str(haplotype):
                lengths.append(pos[1])
        if len(lengths) > 0:
            aligned_het_bases = self.calculate_aligned_het_bases_from_local_alignment(
                    self.local_realignment(lengths[0])
                    )
            self.aligned_het_bases = aligned_het_bases
            for offset in range(0, int(self.allele_length_for_haplotype_i(haplotype))):
                if aligned_het_bases[str(haplotype)][self.gapped_alignment_position + offset]['match']:
                    self.probs_read.append(
                            aligned_het_bases[str(haplotype)][self.gapped_alignment_position + offset][
                                'error_rate_quality_score']
                            )
                    self.total_matches += 1
                elif not aligned_het_bases[str(haplotype)][self.gapped_alignment_position + offset]['match']:
                    self.probs_read.append(
                            float(
                                    aligned_het_bases[str(haplotype)][self.gapped_alignment_position + offset][
                                        'error_rate_quality_score']
                                    ) / 3
                            )
                    self.total_non_matches += 1
            
            return sum(self.probs_read)
    
    @property
    def _generate_gapped_alignment_position(self) -> object:
    
        # reference_start, cigartuples, reference_positions):
        """
        # self.gapped_alignment_positions =
        self._generate_gapped_alignment(self.aligned_segment.reference_start,
        self.aligned_segment.cigartuples, self.reference_positions)

        Returns:
            object:
        """
    
        gapped_alignment_positions = []
        reference_start = self.aligned_segment.reference_start
        cigartuples = self.aligned_segment.cigartuples
        reference_positions: List[int] = [self.reference_position]
        
        i = reference_start  # self.aligned_segment.reference_start
        j = 0
        h = 0
        if not cigartuples:
            return gapped_alignment_positions
        
        cigartuple: tuple
        for cigartuple in cigartuples:  # self.aligned_segment.cigartuples:
            j1 = j
            if h >= len(reference_positions):
                break
            elif cigartuple[0] in [0, 7, 8]:
                i += cigartuple[1]
                j += cigartuple[1]
            elif cigartuple[0] in [1, 4]:
                j += cigartuple[1]
            elif cigartuple[0] in [2, 3]:
                i += cigartuple[1]
            while i > reference_positions[h]:
                if cigartuple[0] in [0, 7, 8]:
                    gapped_alignment_positions.append(j1 + (cigartuple[1] - (i - reference_positions[h])))
                    if h == len(reference_positions) - 1:
                        h += 1
                        break
                elif cigartuple[0] == 2:
                    gapped_alignment_positions.append(j1 + (cigartuple[1] - (i - reference_positions[h])))
                    if h == len(reference_positions) - 1:
                        h += 1
                        break
                if h >= len(reference_positions) - 1:
                    break
                else:
                    h += 1

        if len(gapped_alignment_positions) > 0:
            self._gapped_alignment_position = gapped_alignment_positions[0]
        return gapped_alignment_positions
    
    @property
    def correct_phase(self) -> object:
        """SIMULATED SUBCLASS :returns: :rtype: int"""
        # if not self.aligned_segment.has_tag('oa'):
        #     return
        return self.haplotype
    
    def read_base_quality_error_rate_for_offset(self, position):
        """
        Args:
            position:
        """
        pass


def calculate_favored_haplotype_from_local_alignment(alignments: List[object]) -> int:
    """
    Args:
        alignments (object):
    """
    
    diff: List[Any] = []
    scores: defaultdict[Any, list] = defaultdict(list)
    haplotype: object
    score: object
    for haplotype in alignments:
        scores[alignments[haplotype].score].append(haplotype)
    if len(scores[max(scores.keys())]) == 1:
        haplotype_favored: int = scores[max(scores.keys())][0]
        diff = []
        
        for score in scores.keys():
            if score != max(scores.keys()):
                diff.append(max(scores.keys()) - score)
    
    return haplotype_favored



"""
Helper functions.
"""

import sys, os
import pysam

def is_gzip(filepath):
    """
    Test an input file for (likely) gzip format.
    """
    with open(filepath, 'rb') as test_f:
        return test_f.read(2) == b'\x1f\x8b'

    
def file_test(filepath, file_type=None, allow_gzip=False):
    """
    Test to see if filepath matches file_type given based on some nominal tests.
    """
    if not os.path.isfile(filepath):
        raise Exception('File %s does not exist.' % filepath)

    if file_type == None:
        return True
    else:
        fname = os.path.baseame(filepath)

        # gzip test only
        if file_type in ['gz' or 'gzip']:
            return is_gzip(filepath)

        # See if gzipped file is allowable and test file accordingly.
        if allow_gzip:
            fparts = os.path.splitext(fname)
            if fparts[1] in ['gz', 'gzip']:
                fname = fparts[0]
        else:
            if is_gzip(filepath):
                raise Exception('Gzip format not allowed.')

        # Fastq format test
        if file_type == 'fastq' and (str(fname).lower().endswith('.fastq') or str(fname).lower().endswith('.fq')):
            return True

        
        
    return False



#!/usr/bin/env python

import sys
if sys.version_info > (3, ) and sys.version_info < (3, 7):
    sys.exit("ERROR: LRphase requires Python 3.7 or greater")

#from __future__ import absolute_import
#from __future__ import print_function

from glob import glob
from os.path import basename, splitext
from pathlib import Path

try:
    from setuptools import setup, find_packages
except ImportError:
    from ez_setup import use_setuptools
    use_setuptools()

    
def readme():
    if sys.version_info > (3, ):
        with open(Path(__file__).parent.resolve() / 'README.md', encoding='utf-8') as md:
            return md.read()
    else:
        with open('README.md') as md:
            return md.read()


def main():

    metadata = dict(
        name = 'LRphase',
        version = '0.7.11',
        license = 'MIT',
        description = 'Phasing individual long reads using known haplotype information.',
        description_content_type = 'text/plain',
        long_description = readme(),
        long_description_content_type = 'text/markdown',
        author = 'Greg Farnum',
        author_email = 'gregfar@umich.edu',
        url = 'https://github.com/Boyle-Lab/LRphase.git',
        packages = find_packages('src'),
        package_dir = {'':'src'},
        py_modules = [splitext(basename(path))[0] for path in glob('src/LRphase/*.py')],
        include_package_data = True,
        zip_safe = False,
        classifiers = [
            # complete classifier list: http://pypi.python.org/pypi?%3Aaction=list_classifiers
            'License :: OSI Approved :: MIT License',
            'Programming Language :: Python',
            'Programming Language :: Python :: 3',
            'Programming Language :: Python :: 3.7',
            'Programming Language :: Python :: 3.8',
            'Programming Language :: Python :: 3.9',
            'Programming Language :: Python :: Implementation :: CPython',
            'Programming Language :: Python :: Implementation :: PyPy',
            'Intended Audience :: Science/Research',
            'Natural Language :: English',
            'Operating System :: Microsoft',
            'Operating System :: MacOS',
            'Operating System :: Unix',
            'Programming Language :: Python',
            'Topic :: Documentation',
            'Topic :: Scientific/Engineering :: Bio-Informatics',
            'Topic :: Software Development :: Libraries :: Python Modules',
            'Topic :: Software Development :: Version Control :: Git'
        ],
        project_urls = {
            'Issue Tracker':'https://github.com/Boyle-Lab/LRphase/issues',
        },
        keywords = [
            'long-reads', 'phasing', 'haplotype',
        ],
        python_requires = '>=3.7',
        install_requires = [
            'pysam>=0.16.0.1',
            'biopython>=1.78',
            'pyliftover>=0.4',
            'powerlaw>=1.4.6',
            'numpy>=1.20.1'
            'requests>=2.26.0'
        ],
        extras_require = {
            'mappy':['mappy'],
        },
        setup_requires=[
            'pytest-runner',        
        ],
        entry_points = {
            'console_scripts':[
            'LRphase = LRphase.cli:main',
            ]
        }
    )

    setup(**metadata)


if __name__ == "__main__":
    main()



__version__ = '0.000.1'
try:
    from ._LRphase import longest  # noqa
except ImportError:
    def longest(args):
        """
        Args:
            args:
        """
        return max(args, key=len)


import math
import sys, os
from collections import defaultdict
from pysam import VariantFile, AlignmentFile
#import numpy as np
import pyliftover
from LRphase import PhasableSample, PhaseSet


def true_alignment_match(
    aligned_segment: object,
    needs_liftover: object = False,
    contig: object = None,
    ref_start: object = None,
    ref_end: object = None,
    strand: object = None,
    tag_read: object = False,
    only_output_match_label: object = False,
    only_output_overlap: object = False,
    true_reference_sequence_path: object = None,
    aligned_reference_sequence_path: object = None,
    liftover_converter: object = None,
    chain_file_path: object = None,
    sample: object = None,
    haplotype: object = None,
    reference_liftover_converters: object = None
) -> object:
    """
    Args:
    aligned_segment:
    needs_liftover:
    contig:
    ref_start (object):
    ref_end:
    strand:
    tag_read:
    only_output_match_label:
    only_output_overlap:
    true_reference_sequence_path:
    aligned_reference_sequence_path (object):
    liftover_converter:
    chain_file_path:
    sample:
    haplotype:
    reference_liftover_converters:
    
    Returns:
    
    """
    overlap = 0
    match_label = 'non_match'
    
    if contig is None or ref_start is None or ref_end is None or strand is None:
        contig, ref_start, ref_end, strand = true_alignment_coordinates(
                aligned_segment, tag_read = tag_read, contig = contig, ref_start = ref_start, ref_end = ref_end,
                strand = strand
                )
    
    if needs_liftover:
        if liftover_converter is not None:
            liftover_converter = liftover_converter
    
        elif reference_liftover_converters is not None:
            if aligned_reference_sequence_path is not None:
                if aligned_reference_sequence_path in reference_liftover_converters:
                    if sample in reference_liftover_converters[aligned_reference_sequence_path]:
                        if str(haplotype) in reference_liftover_converters[aligned_reference_sequence_path][sample]:
                            liftover_converter = reference_liftover_converters[aligned_reference_sequence_path][sample][
                                str(haplotype)]
    
        elif chain_file_path is not None:
            if os.path.isfile(chain_file_path):
                liftover_converter = pyliftover.Liftover(chain_file_path)
    
        else:
            sys.stderr.write('could not liftover\n')
            return

        ref_start_lifted = liftover_converter.convert_coordinate(contig, ref_start)
        ref_end_lifted = liftover_converter.convert_coordinate(contig, ref_end)
        
        contig = ref_start_lifted[0][0]
        ref_start = ref_start_lifted[0][1]
        ref_end = ref_end_lifted[0][1]
        strand = ref_start_lifted[0][2]

    """
    sys.stderr.write("True: %s:%s-%s, %s; Observed: %s:%s-%s\n" % (contig, ref_start, ref_end, strand,
                                                                   aligned_segment.reference_name,
                                                                   aligned_segment.reference_start,
                                                                   aligned_segment.reference_end))
    """
    if aligned_segment.reference_name == contig:
        match_label = 'ref_match'
        if int(aligned_segment.reference_start) >= int(ref_start) and int(aligned_segment.reference_start) <= int(
                ref_end
                ):
            match_label = 'mapping_match'
            if int(aligned_segment.reference_end) >= int(ref_end):
                overlap = int(ref_end) - int(aligned_segment.reference_start)
            else:
                overlap = int(aligned_segment.reference_end) - int(aligned_segment.reference_start)
        elif int(aligned_segment.reference_end) >= int(ref_start) and int(aligned_segment.reference_end) <= int(
                ref_end
                ):
            match_label = 'mapping_match'
            if int(ref_start) >= int(aligned_segment.reference_start):
                overlap = int(aligned_segment.reference_end) - int(ref_start)
            else:
                overlap = int(aligned_segment.reference_end) - int(aligned_segment.reference_start)
        elif aligned_segment.reference_start >= (
                int(ref_start) - (int(ref_start) * 0.1)) and aligned_segment.reference_start <= (
                int(ref_end) + (int(ref_end) * 0.1)):
            match_label = 'within_10percent'    
            
    if tag_read:
        aligned_segment.set_tag(
                tag = 'ov', value = str(overlap / (int(ref_end) - int(ref_start))), value_type = 'Z', replace = True
                )
        aligned_segment.set_tag(
                tag = 'OV', value = str(overlap / (aligned_segment.query_alignment_length)), value_type = 'Z',
                replace = True
                )
        aligned_segment.set_tag(tag = 'ml', value = str(match_label), value_type = 'Z', replace = True)

    if only_output_match_label:
        return match_label

    elif only_output_overlap:
        return overlap

    else:
        return match_label, overlap


def true_alignment_coordinates(
        aligned_segment, tag_read = False, contig = None, ref_start = None, ref_end = None, strand = None
        ):
    if '!' in str(aligned_segment.query_name):
        _read_name, _contig, _ref_start, _ref_end, _strand = aligned_segment.query_name.split('!')
        if contig is None:
            contig = _contig.split(':')[0]
        if ref_start is None:
            ref_start = _ref_start
        if ref_end is None:
            ref_end = _ref_end
        if strand is None:
            strand = _strand
    
    if tag_read:
        aligned_segment.set_tag(tag = 'st', value = str(ref_start), value_type = 'Z', replace = True)
        aligned_segment.set_tag(tag = 'lo', value = str(contig), value_type = 'Z', replace = True)
        aligned_segment.set_tag(tag = 'en', value = str(ref_end), value_type = 'Z', replace = True)
        aligned_segment.set_tag(tag = 'sd', value = str(strand), value_type = 'Z', replace = True)
    
    return contig, ref_start, ref_end, strand


def true_read_origin(
        aligned_segment, tag_read = False, sample = None, haplotype = None, contig = None, ref_start = None,
        ref_end = None, strand = None
        ):
    if '!' in str(aligned_segment.query_name):
        _ref_start: object
        _read_name, _contig, _ref_start, _ref_end, _strand = aligned_segment.query_name.split('!')
        if contig is None:
            contig = _contig
        if ref_start is None:
            ref_start = _ref_start
        if ref_end is None:
            ref_end = _ref_end
        if strand is None:
            strand = _strand
        
        if '__' in _read_name:
            sample, haplotype, name_contig_numeric = _read_name.split('__')
    
    if tag_read:        
        aligned_segment.set_tag(tag = 'sm', value = str(sample), value_type = 'Z', replace = True)
        aligned_segment.set_tag(tag = 'ha', value = str(haplotype), value_type = 'Z', replace = True)
    
    return sample, haplotype


def alignment_type(aligned_segment, tag_read = False):
    alignment_type = 'None'
    if aligned_segment.is_unmapped:
        alignment_type = 'unmapped'
    elif aligned_segment.is_secondary:
        alignment_type = 'secondary'
    elif aligned_segment.is_supplementary:
        alignment_type = 'supplementary'
    else:
        alignment_type = 'mapped'
    if tag_read:
        aligned_segment.set_tag(tag = 'al', value = str(alignment_type), value_type = 'Z', replace = True)
    
    return alignment_type


class PhasedRead:
    """

    """
    
    def __init__(
            self, aligned_segment, phasable_sample = None, vcf_file = None, sample = None,
            haplotype = None, ignore_phase_sets = False, error_model = 0,
            error_rate_threshold = 0.01,
            prior_probabilities = None, bam_file_header = None, bam_file_template = None,
            output_file_path = None, liftover_converters = None,
            multinomial_correction = True, auto_phase = True, evaluate_alignment = True,
            evaluate_true_alignment = False, aligned_reference_sequence_path = None,
            powlaw_alpha = 4.5, powlaw_xmin = 2.0
    ):
        self.output_file_path = None
        self.liftover_converters = None
        self.aligned_segment = aligned_segment
        self.aligned_reference_sequence_path = None
        self.needs_liftover = False
        self.true_reference_sequence_path = None
        
        if isinstance(phasable_sample, PhasableSample.PhasableSample):
            self.vcf_file=phasable_sample.vcf_file
            self.sample=phasable_sample.sample
            if phasable_sample.simulate_haplotypes:
                if output_file_path is None:
                    self.output_file_path = phasable_sample.output_directory
                self.liftover_converters = phasable_sample.haplotype_liftover_converters
                self.aligned_reference_sequence_path = phasable_sample.reference_sequence_path
                self.needs_liftover = True
                self.evaluate_true_alignment = True
        
        else:
            self.vcf_file = vcf_file
            self.sample = sample
        
        self.haplotype = haplotype
        self.ignore_phase_sets = ignore_phase_sets

        # Attributes related to error-rate thresholding
        self.error_model = error_model
        self.multinomial_correction = multinomial_correction
        self.error_rate_threshold = error_rate_threshold
        self.prior_probabilities = prior_probabilities
        self.powlaw_alpha = powlaw_alpha
        self.powlaw_xmin = powlaw_xmin

        # self._get_alignment_label()
        if self.liftover_converters is None:
            self.liftover_converters = liftover_converters
        self.auto_phase = auto_phase
        self._Phase_Set_max = None
        self.PhaseSets = []
        self.evaluate_alignment = evaluate_alignment
        self.evaluate_true_alignment = evaluate_true_alignment
        
        if output_file_path:
            self.output_file_path = output_file_path
        else:
            self.output_file_path = '%s_phase_tagged.bam' % self.sample
            
        if bam_file_header:
            self.bam_file_header = bam_file_header
        else:
            self.bam_file_header = None

        if bam_file_template:
            self.bam_file_template = AlignmentFile(bam_file_template, 'rb')
        else:
            self.bam_file_template = None
            
        if self.evaluate_alignment:
            alignment_type(self.aligned_segment, tag_read = True)
            #self._evaluate_alignment()
        if self.evaluate_true_alignment:
            true_alignment_match(
                self.aligned_segment, 
                needs_liftover = self.needs_liftover, 
                contig = None,
                ref_start = None,
                ref_end = None,
                strand = None,
                tag_read = True,
                only_output_match_label = False,
                only_output_overlap = False,
                true_reference_sequence_path = self.true_reference_sequence_path,
                aligned_reference_sequence_path = self.aligned_reference_sequence_path,
                liftover_converter = self.liftover_converters,
                chain_file_path = None,
                sample = self.sample,
                haplotype = self.haplotype,
                reference_liftover_converters = None
            )
            true_read_origin(
                self.aligned_segment,
                tag_read = True,
                sample = self.sample,
                haplotype = self.haplotype,
                contig = None,
                ref_start = None,
                ref_end = None,
                strand = None
            )
        
        if self.auto_phase:
            self.phase_read(
                    error_model = self.error_model, error_rate_threshold = self.error_rate_threshold,
                    multinomial_correction = self.multinomial_correction,
                    )
    
    def __repr__(self):
        return f'Read Name: {self.query_name}, Is phased: {self.is_phased}, Favors single phase: {self.one_phase_is_favored}\nRead Length: {self.aligned_segment.query_length}\nAlignment Type: {self.alignment_type}\nPrimary Alignment Length: {self.aligned_segment.query_alignment_length}'
    
    def __iter__(self):
        self.PhaseSets = self._find_PhaseSets(
                self.error_model, self.error_rate_threshold, self.prior_probabilities,
                liftover_converters = self.liftover_converters
                )
        self.PhaseSets_processed_count = 0
        # self.alignment_files_read_counts = [bam_file.mapped+bam_file.unmapped for bam_file in self.bam_files]
        return self
    
    def __next__(self):
        if self.PhaseSets_processed_count < len(self.PhaseSets):
            PhaseSet = self.PhaseSets[self.PhaseSets_processed_count]
            self.PhaseSets_processed_count += 1
            return PhaseSet
        else:
            raise StopIteration()
    
    def _find_PhaseSets(self, error_model, error_rate_threshold, prior_probabilities, liftover_converters = None):
        
        if liftover_converters is None:
            liftover_converters = self.liftover_converters
        
        al_type = alignment_type(self.aligned_segment, tag_read=True)
        
        if self.aligned_segment.is_unmapped:
            # Return an empty phaseSet
            return []
        
        if self.aligned_segment.get_tag('al') == 'aligned_to_contig_not_in_vcf' or self.aligned_segment.is_unmapped:
            return 'aligned_to_contig_not_in_vcf'
        
        variants_phase_sets = self._fetch_phased_variants(
                self.aligned_segment, self.vcf_file, self.sample, self.ignore_phase_sets
                )
        
        PhaseSets = []
        if len(variants_phase_sets) > 0:
            if max([len(variants_phase_sets[str(phase_set)]) for phase_set in variants_phase_sets]) > 0:
                for phase_set in variants_phase_sets:
                    PhaseSets.append(
                        PhaseSet.PhaseSet(
                            phase_set,
                            self.aligned_segment,
                            variants_phase_sets[str(phase_set)],
                            str(self.sample),
                            error_model = error_model,
                            error_rate_threshold = error_rate_threshold,
                            prior_probabilities = prior_probabilities,
                            liftover_converters = liftover_converters,
                            powlaw_alpha = self.powlaw_alpha,
                            powlaw_xmin = self.powlaw_xmin
                        )
                    )
        
        return PhaseSets
    
    def phase_read(
        self, 
        error_model = None,
        error_rate_threshold = None, 
        prior_probabilities = None,
        multinomial_correction = None
    ):
        '''#self.Phase_Set_max, self.phase, self.phase_set_name, self.log_likelihood_ratio, self.max_phase,
        self.ploidy_phase_set = self._select_best_phasing(self._find_PhaseSets()) '''
        if error_model is None:
            error_model = self.error_model
        if error_rate_threshold is None:
            error_rate_threshold = self.error_rate_threshold
        if multinomial_correction is None:
            multinomial_correction = True
        
        #if evaluate_alignment:
            #self._evaluate_alignment()
        #if evaluate_true_alignment:
        #    if self.aligned_segment.has_tag('oa'):
        #        match_label = self._evaluate_true_alignment
        
        if not self.aligned_segment.has_tag('al'):
            alignment_type(self.aligned_segment,tag_read=True)
        
        #if not self.aligned_segment.has_tag('ml'):
        #    if self.aligned_segment.has_tag('oa'):
        #        match_label = self._evaluate_true_alignment
        
        #_Phase_Set_max, phase, phase_set_name, log_likelihood_ratio, max_phase = self._select_best_phasing(
        _Phase_Set_max = self._select_best_phasing(
            self._find_PhaseSets(error_model, error_rate_threshold, prior_probabilities,liftover_converters = self.liftover_converters),
            error_model,
            error_rate_threshold,
            prior_probabilities,
            multinomial_correction
        )
        
        self._Phase_Set_max = _Phase_Set_max
        
        self.aligned_segment.set_tag(tag = 'PS', value = str(self.phase_set_name), value_type='Z', replace=True)
        self.aligned_segment.set_tag(tag = 'HP', value = str(self.phase), value_type='Z', replace=True)
        self.aligned_segment.set_tag(tag = 'PC', value = str(self.log_likelihood_ratio), value_type='Z', replace=True)
        self.aligned_segment.set_tag(tag = 'py', value = str(self.ploidy_phase_set), value_type='Z', replace=True)
        self.aligned_segment.set_tag(tag = 'HS', value = str(self._Phase_Set_max.total_hets), value_type='Z', replace=True)
        
        return self
    
    def _select_best_phasing(
            self, PhaseSets, error_model, error_rate_threshold, prior_probabilities = None,
            multinomial_correction = True
            ):
        '''self.Phase_Set_max, self.phase, self.phase_set_name, self.log_likelihood_ratio, self.max_phase =
        self._select_best_phasing(self.PhaseSets) '''
        self.PhaseSets = PhaseSets
        phasable = {}
        if isinstance(PhaseSets, str):
            return PhaseSet.PhaseSet(None, None, None, None)
        
        if len(PhaseSets) > 0:
            for Phase_Set in PhaseSets:
                Phase_Set.solve_phase(error_model, error_rate_threshold, prior_probabilities, multinomial_correction)
                if isinstance(Phase_Set.max_log_likelihood_ratio, float):
                    phasable[Phase_Set.max_log_likelihood_ratio] = Phase_Set
        
        if len(phasable) > 0:
            Phase_Set_max = phasable[max(phasable.keys())]
        else:
            Phase_Set_max = PhaseSet.PhaseSet(None, self.aligned_segment, [], str(self.sample))
        
        return Phase_Set_max#, phase, phase_set_name, log_likelihood_ratio, max_phase
    
    # @staticmethod
    def _fetch_phased_variants(self, aligned_segment, vcf_file, sample, ignore_phase_sets):
        '''self.variants_phase_sets = self._fetch_phased_variants(self.aligned_segment, self.vcf_file, self.sample,
        self.ignore_phase_sets) '''
        variants_phase_sets = defaultdict(list)

        vcf_in = VariantFile(vcf_file)

        # VariantFile.fetch() will cause program termination if it encounters a contig
        # that is not found in the given VCF file. Therefore we need to encapsulate this
        # in a try block to prevent premature termination. Aligned segments that map to
        # such contigs will be skipped.
        try:
            vcf_recs = vcf_in.fetch(aligned_segment.reference_name, aligned_segment.reference_start, aligned_segment.reference_end)
            # Note that coordinates supplied in this way are assumed zero-based by pysam (see source code, https://github.com/pysam-developers/pysam/blob/master/pysam/libcbcf.pyx, line 4366. We should verify these are zero-based. If not, use pysam chr:start-end string -- these are treated as one-based. This may not be a huge deal, since only variants coinciding with the start/end of the fragment will be affected.
            
            for vcf_record in vcf_recs:
                if vcf_record.samples[str(sample)].phased and max(vcf_record.samples[str(sample)].allele_indices) > min(
                        vcf_record.samples[str(sample)].allele_indices
                        ):
                    if ignore_phase_sets:
                        variants_phase_sets['ignore_phase_sets'].append(vcf_record)
                    else:
                        variants_phase_sets[str(vcf_record.samples[str(sample)]['PS'])].append(vcf_record)
        except:
            # Do some error handling...
            pass
        
        vcf_in.close()
        return variants_phase_sets

    def write_to_bam(self, output_file_path_bam=None, output_bam_pysam=None, header=None, template=None):
        """
        Args:
            output_file_path_bam: Path to output bam file.
            output_bam_pysam: Pysam AlignmentFile object invoked with 'wb' flags.
            header: Pysam AlignmentHeader object or dictionary.
            template: Pysam AlignmentFile object invoked with 'rb' flags from which bam header will be extracted.
        Returns: Nothing
        """

        if header and template:
            sys.stderr.write('WARNING: Both header and template given for bam output. Template will be used by default.')        
        _bam_file_header = None
        if header:
            _bam_file_header = header
        elif self.bam_file_header:
            _bam_file_header = self.bam_file_header

        _bam_file_template = None
        if template:
            _bam_file_template = template
        elif self.bam_file_template:
            _bam_file_template = self.bam_file_template
            
        if not _bam_file_header and not _bam_file_template and not output_bam_pysam:
            raise Exception('Bam output requires header, template, or pysam AlignmentFile object (output_bam_pysam). None given.')
            
        if output_bam_pysam:
            output_bam = output_bam_pysam
        else:
            if output_file_path_bam:
                _output_file_path = output_file_path_bam
            else:
                _output_file_path = self.output_file_path
            if _bam_file_template:
                output_bam = AlignmentFile(_output_file_path, 'wb', template = _bam_file_template)
            else:
                output_bam = AlignmentFile(_output_file_path, 'wb', header = _bam_file_header)

        # self.aligned_segment.set_tag(tag = 'RG', value = str(RG_info[1]), value_type='Z', replace=True)
        output_bam.write(self.aligned_segment)

        return

    
    @property
    def is_phased_correctly(self):
        '''
        '''
        if self.Phase_Set_max is None:
            return
        elif self.Phase_Set_max.phase == 'Nonphasable':
            if self.is_phased:
                return False
            else:
                return
        else:
            if self.is_phased:
                if self.aligned_segment.has_tag('ha'):
                    return str(self.phase) == str(self.aligned_segment.get_tag('ha'))
    
    @property
    def read_name(self):
        '''
        '''
        return str(self.aligned_segment.query_name)
    
    @property
    def query_name(self):
        '''
        '''
        return str(self.aligned_segment.query_name)
    
    @property
    def alignment_type(self):
        '''
        '''
        if self.aligned_segment.has_tag('al'):
            return str(self.aligned_segment.get_tag('al'))
    
    @property
    def fraction_of_read_overlapping_true_alignment(self):
        '''
        '''
        if self.aligned_segment.has_tag('ov'):
            return str(self.aligned_segment.get_tag('ov'))
    
    @property
    def alignment_is_unmapped(self):
        '''
        '''
        return self.aligned_segment.is_unmapped
    
    @property
    def alignment_is_supplementary(self):
        '''
        '''
        return self.aligned_segment.is_supplementary
    
    @property
    def alignment_is_secondary(self):
        '''
        '''
        return self.aligned_segment.is_secondary
    
    @property
    def is_aligned_to_contig_not_in_vcf(self):
        '''
        '''
        return self.aligned_segment.reference_name not in list(self.vcf_file.index)
    
    @property
    def alignment_is_reverse(self):
        '''
        true if aligned_segment is mapped to reverse strand
        '''
        return self.aligned_segment.is_reverse
    
    @property
    def alignment_is_mapped(self):
        '''
        '''
        return self.aligned_segment.is_unmapped == False
    
    @property
    def alignment_is_primary(self):
        '''
        '''
        if self.aligned_segment.has_tag('tp'):
            return str(self.aligned_segment.get_tag('tp')) == 'P'
        # return self.aligned_segment.is_unmapped == False
    
    @property
    def validation_using_true_alignment_label(self):
        '''
        '''
        if self.aligned_segment.has_tag('ml'):
            return str(self.aligned_segment.get_tag('ml'))
    
    @property
    def matches_true_alignment(self):
        '''
        '''
        if self.aligned_segment.has_tag('ml'):
            return str(self.aligned_segment.get_tag('ml')) == 'mapping_match'
    
    @property
    def mapped_within_10percent_of_true_alignment(self):
        '''
        '''
        if self.aligned_segment.has_tag('ml'):
            return str(self.aligned_segment.get_tag('ml')) == 'within_10percent'
    
    @property
    def does_not_overlap_true_alignment_but_does_align_to_true_contig(self):
        '''
        '''
        if self.aligned_segment.has_tag('ml'):
            return str(self.aligned_segment.get_tag('ml')) == 'ref_match'
    
    @property
    def does_not_match_true_alignment(self):
        '''
        '''
        if self.aligned_segment.has_tag('ml'):
            return str(self.aligned_segment.get_tag('ml')) == 'non_match'
    
    @property
    def alignment_length_fraction_of_total_read_length(self):
        '''
        aligned_segment.query_alignment_length/aligned_segment.query_length
        returns 0 if unmapped
        '''
        # self.aligned_segment.set_tag(tag = 'xx', value = str(self.aligned_segment.query_alignment_length/self.aligned_segment.query_length),
        # value_type='Z', replace=True)
        if self.aligned_segment.has_tag('al'):
            if str(self.aligned_segment.get_tag('al')) == 'unmapped_reads':
                return 0
            return self.aligned_segment.query_alignment_length / self.aligned_segment.query_length
    
    @property
    def fraction_of_alignment_length_overlapping_true_alignment(self):
        '''
        '''
        if self.aligned_segment.has_tag('OV'):
            return str(self.aligned_segment.get_tag('OV'))
    
    @property
    def favored_phase_is_correct(self):
        '''
        '''
        if self.Phase_Set_max is None:
            return
        elif self.Phase_Set_max == 'Nonphasable':
            return False
        else:
            if self.one_phase_is_favored:
                if self.aligned_segment.has_tag('ha'):
                    return str(self.max_phase) == str(self.aligned_segment.get_tag('ha'))
    
    @property
    def is_phased_incorrectly(self):
        '''
        '''
        if self.Phase_Set_max is None:
            return
        elif self.Phase_Set_max == 'Nonphasable':
            return False
        else:
            if self.is_phased:
                if self.aligned_segment.has_tag('ha'):
                    return str(self.phase) != str(self.aligned_segment.get_tag('ha'))
    
    @property
    def favored_phase_is_incorrect(self):
        '''
        '''
        if self.Phase_Set_max is None:
            return
        elif self.Phase_Set_max == 'Nonphasable':
            return False
        else:
            if self.one_phase_is_favored:
                if self.aligned_segment.has_tag('ha'):
                    return str(self.max_phase) != str(self.aligned_segment.get_tag('ha'))
    
    @property
    def is_Nonphasable(self):
        '''
        Returns True if this aligned_segment was assigned Nonphasable for not having any heterozygous positions
        '''
        if self.Phase_Set_max is None:
            return
        elif self.Phase_Set_max == 'Nonphasable':
            return True
        else:
            return self.phase == 'Nonphasable'
    
    @property
    def is_Unphased(self):
        '''
        Returns True if this aligned_segment was left unassigned (Unphased) because it did not have a log likelihood ratio (LR)
        above the threshold chosen for this experiment.
        '''
        if self.Phase_Set_max is None:
            return
        elif self.Phase_Set_max == 'Nonphasable':
            return False
        else:
            return self.phase == 'Unphased'
    
    @property
    def is_phased(self):
        '''
        Returns True if this aligned_segment was assigned to a haplotype based on its log likelihood ratio (LR) being above the
        threshold chosen for this experiment.
        '''
        if self.Phase_Set_max is None:
            return
        elif self.Phase_Set_max == 'Nonphasable':
            return False
        else:
            if self.phase == 'Unphased' or self.phase == 'Nonphasable':
                return False
            else:
                return True
        return False
    
    @property
    def one_phase_is_favored(self):
        '''
        Returns True if this aligned_segment was assigned to a haplotype based on its log likelihood ratio (LR) being above the
        threshold chosen for this experiment.
        '''
        if self.Phase_Set_max is None:
            return
        elif self.Phase_Set_max.phase == 'Nonphasable':
            return False
        else:
            if int(self.max_phase) > 0:
                one_phase_is_favored = True
            elif int(self.max_phase) == 0:
                one_phase_is_favored = False
            else:
                return
            return one_phase_is_favored
    
    def is_assigned_to_haplotype_i(self, haplotype):
        
        '''Returns True if this aligned_segment was was assigned to haplotype i. (ie: the index of the haplotype in the allele_indices A|T )
        '''
        if self.Phase_Set_max is None:
            return
        elif self.Phase_Set_max == 'Nonphasable':
            return False
        else:
            return str(self.phase) == str(haplotype)
    
    @property
    def overlaps_multiple_phase_sets(self):
        '''
        This aligned_segment overlapped heterozygous positions that belonged to at least two different phase sets (independently
        phased blocks of sequence are only phased within themselves and one must be chosen)
        '''
        if self.Phase_Set_max is None:
            return
        elif self.Phase_Set_max == 'Nonphasable':
            return False
        if len(self.PhaseSets) > 1:
            overlaps_multiple_phase_sets = True
        else:
            overlaps_multiple_phase_sets = False
        return overlaps_multiple_phase_sets
    
    @property
    def Phase_Set_max(self):
        '''
        '''
        if self._Phase_Set_max is None:
            return
        elif self._Phase_Set_max == 'Nonphasable':
            return 'Nonphasable'
        else:
            return self._Phase_Set_max
    
    @property
    def phase(self):
        '''
        '''
        if self.Phase_Set_max is None:
            return
        elif self.Phase_Set_max.phase == 'Nonphasable':
            return 'Nonphasable'
        else:
            return self.Phase_Set_max.phase
    
    @property
    def max_phase(self):
        '''
        The number of distinct haplotypes at this position. Our model assumes that ploidy represents the ground truth
        molecular copy number of the population from which these reads were samples.
        '''
        if self.Phase_Set_max is None:
            return
        elif self.Phase_Set_max.phase == 'Nonphasable':
            return 'Nonphasable'
        else:
            return self.Phase_Set_max.max_phase
    
    @property
    def phase_set_name(self):
        '''
        The number of distinct haplotypes at this position. Our model assumes that ploidy represents the ground truth
        molecular copy number of the population from which these reads were samples.
        '''
        if self.Phase_Set_max is None:
            return
        elif self.Phase_Set_max.phase == 'Nonphasable':
            return 'Nonphasable'
        else:
            return self.Phase_Set_max.phase_set
    
    @property
    def phasing_error_rate(self):
        if self.phase == 'Nonphasable':
            return 0.50
        if self.log_likelihood_ratio > 0:
            p = self.powlaw_modified(self.log_likelihood_ratio,
                                     a = self.powlaw_alpha,
                                     xmin = self.powlaw_xmin)
            pl = 1 - (1 / float(self.ploidy_phase_set))
            if p <= pl:
                return p
            else:
                return pl
    
    def powlaw_modified(self, x, a = 4.5, xmin = 2):
        return ((a - 1) / xmin) * math.pow(x / xmin, -1 * a)
    
    @property
    def log_likelihood_ratio(self):
        '''
        The number of distinct haplotypes at this position. Our model assumes that ploidy represents the ground truth
        molecular copy number of the population from which these reads were samples.
        '''
        if self.Phase_Set_max is None:
            return
        elif self.Phase_Set_max.phase == 'Nonphasable':
            return 'Nonphasable'
        else:
            return self.Phase_Set_max.max_log_likelihood_ratio
    
    @property
    def ploidy_phase_set(self):
        '''
        The number of distinct haplotypes at this position. Our model assumes that ploidy represents the ground truth
        molecular copy number of the population from which these reads were samples.
        '''
        if self.Phase_Set_max is None:
            return
        elif self.Phase_Set_max.phase == 'Nonphasable':
            return 'Nonphasable'
        else:
            return self.Phase_Set_max.ploidy_phase_set
    
    @property
    def error_model_used(self):
        '''
        self.aligned_segment.set_tag(tag = 'em', value = str(self.error_model), value_type='Z', replace=True)

        '''
        if self.Phase_Set_max is None:
            return
        elif self.Phase_Set_max.phase == 'Nonphasable':
            return 'Nonphasable'
        else:
            return self.Phase_Set_max.error_model_used
    
    @property
    def multinomial_correction_used(self):
        '''
        self.aligned_segment.set_tag(tag = 'em', value = str(self.error_model), value_type='Z', replace=True)

        '''
        if self.Phase_Set_max is None:
            return
        elif self.Phase_Set_max.phase == 'Nonphasable':
            return False
        else:
            return self.Phase_Set_max.multinomial_correction
    
    @property
    def error_rate_threshold_used(self):
        '''

        self.aligned_segment.set_tag(tag = 'et', value = str(self.error_rate_threshold), value_type='Z', replace=True)
        '''
        if self.Phase_Set_max is None:
            return
        elif self.Phase_Set_max.phase == 'Nonphasable':
            return 'Nonphasable'
        else:
            return self.Phase_Set_max.error_rate_threshold_used
    
    @property
    def prior_probabilities_used(self):
        '''

        self.aligned_segment.set_tag(tag = 'et', value = str(self.error_rate_threshold), value_type='Z', replace=True)
        '''
        if self.Phase_Set_max is None:
            return
        elif self.Phase_Set_max.phase == 'Nonphasable':
            return 'Nonphasable'
        else:
            return self.Phase_Set_max.prior_probabilities_used
    
    @property
    def error_rate_average_het_sites_used(self):
        '''

        self.aligned_segment.set_tag(tag = 'er', value = str(Phase_Set_max.error_rate_average_het_sites_used), value_type='Z', replace=True)
        '''
        if self.Phase_Set_max is None:
            return
        elif self.Phase_Set_max.phase == 'Nonphasable':
            return 'Nonphasable'
        else:
            return self.Phase_Set_max.error_rate_average_het_sites
    
    @property
    def per_base_mismatch_rate_used(self):
        '''

        self.aligned_segment.set_tag(tag = 'er', value = str(Phase_Set_max.error_rate_average_het_sites_used), value_type='Z', replace=True)
        '''
        if self.Phase_Set_max is None:
            return
        elif self.Phase_Set_max.phase == 'Nonphasable':
            return 'Nonphasable'
        else:
            return self.Phase_Set_max.per_base_mismatch_rate
    
    @property
    def total_hets_analyzed_favored_haplotype(self):
        '''
        Returns the total number of heterozygous positions that were included in the phasing evaluation for this aligned_segment
        that resulted in the most favorable likelihood (ie: positions containing heterozygous SNVs that align to the
        aligned_segment of interest.)
        '''
        if self.Phase_Set_max is None:
            return
        elif self.Phase_Set_max.phase == 'Nonphasable':
            return 'Nonphasable'
        else:
            return self.Phase_Set_max.total_hets_analyzed
    
    @property
    def total_hets_favored_haplotype(self):
        '''
        Returns the total number of heterozygous positions that overlapped this aligned_segment before filtering the sites with
        '-'. This set of het sites was the most favorable phase set that resulted in the most favorable likelihood (
        ie: positions containing heterozygous SNVs that align to the aligned_segment of interest.)
        '''
        if self.Phase_Set_max is None:
            return
        elif self.Phase_Set_max.phase == 'Nonphasable':
            return 'Nonphasable'
        else:
            return self.Phase_Set_max.total_hets
    
    @property
    def total_matches_to_favored_haplotype(self):
        '''
        Returns the total number of matches to the favored haplotype sequence at heterozygous positions that
        overlapped this aligned_segment. This phased set of het sites resulted in the most favorable likelihood for this aligned_segment.
        '''
        if self.Phase_Set_max is None:
            return
        elif self.Phase_Set_max.phase == 'Nonphasable':
            return 'Nonphasable'
        else:
            return self.Phase_Set_max.matches[self.Phase_Set_max.max_phase - 1]
    
    @property
    def total_non_matches_to_favored_haplotype(self):
        '''
        Returns the total number of non-matches to the favored haplotype sequence at heterozygous positions that
        overlapped this aligned_segment. This phased set of het sites resulted in the most favorable likelihood for this aligned_segment.
        '''
        if self.Phase_Set_max is None:
            return
        elif self.Phase_Set_max.phase == 'Nonphasable':
            return 'Nonphasable'
        else:
            return self.Phase_Set_max.non_matches[self.Phase_Set_max.max_phase - 1]
    
    def prior_probability_for_haplotype_i(self, i):
        '''
        self.aligned_segment.set_tag(tag = 'pr', value = str(self.prior_probabilities), value_type='Z', replace=True)


        prior_probability_haplotype_i = P(haplotype_i)
        '''
        if self.Phase_Set_max is None:
            return
        elif self.Phase_Set_max.phase == 'Nonphasable':
            return 'Nonphasable'
        else:
            return self.Phase_Set_max.prior_probabilities[i - 1]
    
    @property
    def prior_probability_for_favored_haplotype(self):
        '''
        self.aligned_segment.set_tag(tag = 'pr', value = str(self.prior_probabilities), value_type='Z', replace=True)


        prior_probability_haplotype_i = P(haplotype_i)
        '''
        if self.Phase_Set_max is None:
            return
        elif self.Phase_Set_max.phase == 'Nonphasable':
            return 'Nonphasable'
        else:
            return self.Phase_Set_max.prior_probabilities[self.max_phase - 1]
    
    def calculate_log_likelihood_read_given_haplotype_i(self, i):
        """

        Args:
            i:

        Returns:

        """
        if self.Phase_Set_max is None:
            return
        elif self.Phase_Set_max.phase == 'Nonphasable':
            return 'Nonphasable'
        else:
            return self.Phase_Set_max.log_probability_read_given_haplotype_i[i - 1]
    
    def calculate_likelihood_of_read_given_haplotype_i(self, i):
        """

        Args:
            i:

        Returns:

        """
        if self.Phase_Set_max is None:
            return
        elif self.Phase_Set_max.phase == 'Nonphasable':
            return 'Nonphasable'
        else:
            return 10**self.Phase_Set_max.log_probability_read_given_haplotype_i[i - 1]
    
    @property
    def calculate_log_likelihood_read_given_favored_haplotype(self):
        """

        Returns:

        """
        if self.Phase_Set_max is None:
            return
        elif self.Phase_Set_max.phase == 'Nonphasable':
            return 'Nonphasable'
        else:
            return self.Phase_Set_max.log_probability_read_given_haplotype_i[self.max_phase - 1]
    
    @property
    def calculate_likelihood_of_read_given_favored_haplotype(self):
        """

        Returns:

        """
        if self.Phase_Set_max is None:
            return
        elif self.Phase_Set_max.phase == 'Nonphasable':
            return 'Nonphasable'
        else:
            return 10**self.Phase_Set_max.log_probability_read_given_haplotype_i[self.max_phase - 1]
    
    @property
    def bayes_numerator_for_favored_haplotype(self):
        '''
        P(aligned_segment|haplotype_i)P(haplotype_i)
        Bayes numerator. likelihood * prior
        '''
        if self.Phase_Set_max is None:
            return
        elif self.Phase_Set_max.phase == 'Nonphasable':
            return 'Nonphasable'
        else:
            return self.calculate_likelihood_of_read_given_haplotype_i(
                    self.max_phase - 1
                    ) * self.prior_probability_for_haplotype_i(self.max_phase - 1)
        # if self.likelihood_for_haplotype_i(i) and self.prior_probability_for_haplotype_i(i):
    
    @property
    def log_likelihood_ratio_for_favored_haplotype_vs_not_favored_haplotype(self):
        '''

        '''
        if self.Phase_Set_max is None:
            return
        elif self.Phase_Set_max.phase == 'Nonphasable':
            return 'Nonphasable'
        else:
            if self.total_probability_of_read_given_haplotypes and self.bayes_numerator_for_haplotype_i(
                    self.max_phase - 1
                    ):
                return math.log10(self.bayes_numerator_for_favored_haplotype) - math.log10(
                        sum(
                                [self.bayes_numerator_for_haplotype_i(_i - 1) for _i in
                                 range(1, self.ploidy_phase_set + 1) if
                                 _i - 1 != self.max_phase - 1]
                                )
                        )
    
    @property
    def posterior_probability_for_favored_haplotype(self):
        '''
        # posterior_probability_haplotype_i = P(haplotype_i|aligned_segment)

        # aligned_segment.set_tag(tag = 'fp', value = str(phased_read.posterior_probability_for_haplotype_i(c)), value_type='Z',
        replace=True) # aligned_segment.set_tag(tag = 'pp', value = str(phased_read.posterior_probability_for_haplotype_i(int(
        aligned_segment.get_tag('oa')))), value_type='Z', replace=True)


        '''
        if self.Phase_Set_max is None:
            return
        elif self.Phase_Set_max.phase == 'Nonphasable':
            return 'Nonphasable'
        else:
            if self.total_probability_of_read_given_haplotypes and self.bayes_numerator_for_haplotype_i(
                    self.max_phase - 1
                    ):
                return self.bayes_numerator_for_haplotype_i(
                        self.max_phase - 1
                        ) / self.total_probability_of_read_given_haplotypes
    
    @property
    def log_posterior_probability_for_favored_haplotype(self):
        '''
        posterior_probability_haplotype_i = P(haplotype_i|aligned_segment)
        '''
        if self.Phase_Set_max is None:
            return
        elif self.Phase_Set_max.phase == 'Nonphasable':
            return 'Nonphasable'
        else:
            if self.total_probability_of_read_given_haplotypes and self.bayes_numerator_for_haplotype_i(
                    self.max_phase - 1
                    ):
                return math.log10(self.bayes_numerator_for_haplotype_i(self.max_phase - 1)) - math.log10(
                        self.total_probability_of_read_given_haplotypes
                        )
    
    def bayes_numerator_for_haplotype_i(self, i):
        '''
        P(aligned_segment|haplotype_i)P(haplotype_i)
        Bayes numerator. likelihood * prior
        '''
        if self.Phase_Set_max is None:
            return
        elif self.Phase_Set_max.phase == 'Nonphasable':
            return 'Nonphasable'
        else:
            return self.calculate_likelihood_of_read_given_haplotype_i(i - 1) * self.prior_probability_for_haplotype_i(
                    i - 1
                    )
        # if self.likelihood_for_haplotype_i(i) and self.prior_probability_for_haplotype_i(i):
    
    def log_likelihood_ratio_for_haplotype_i_vs_not_haplotype_i(self, i):
        '''

        '''
        if self.Phase_Set_max is None:
            return
        elif self.Phase_Set_max.phase == 'Nonphasable':
            return 'Nonphasable'
        else:
            if self.total_probability_of_read_given_haplotypes and self.bayes_numerator_for_haplotype_i(i - 1):
                return math.log10(self.bayes_numerator_for_haplotype_i(i - 1)) - math.log10(
                        sum(
                                [self.bayes_numerator_for_haplotype_i(_i - 1) for _i in
                                 range(1, self.ploidy_phase_set + 1) if
                                 _i - 1 != i - 1]
                                )
                        )
    
    def posterior_probability_for_haplotype_i(self, i):
        '''
        # posterior_probability_haplotype_i = P(haplotype_i|aligned_segment)

        # aligned_segment.set_tag(tag = 'fp', value = str(phased_read.posterior_probability_for_haplotype_i(c)), value_type='Z',
        replace=True) # aligned_segment.set_tag(tag = 'pp', value = str(phased_read.posterior_probability_for_haplotype_i(int(
        aligned_segment.get_tag('oa')))), value_type='Z', replace=True)


        '''
        if self.Phase_Set_max is None:
            return
        elif self.Phase_Set_max.phase == 'Nonphasable':
            return 'Nonphasable'
        else:
            if self.total_probability_of_read_given_haplotypes and self.bayes_numerator_for_haplotype_i(i - 1):
                return self.bayes_numerator_for_haplotype_i(i - 1) / self.total_probability_of_read_given_haplotypes
    
    def log_posterior_probability_for_haplotype_i(self, i):
        '''
        posterior_probability_haplotype_i = P(haplotype_i|aligned_segment)
        '''
        if self.Phase_Set_max is None:
            return
        elif self.Phase_Set_max.phase == 'Nonphasable':
            return 'Nonphasable'
        else:
            if self.total_probability_of_read_given_haplotypes and self.bayes_numerator_for_haplotype_i(i - 1):
                return math.log10(self.bayes_numerator_for_haplotype_i(i - 1)) - math.log10(
                        self.total_probability_of_read_given_haplotypes
                        )
    
    @property
    def total_probability_of_read_given_haplotypes(self):
        '''
        total_probability_of_read_given_haplotypes = sum(P(aligned_segment|haplotype_i)P(haplotype_i)) for i = 1,...,ploidy Our
        model assumes that the haplotypes listed in the VCF file are the ground truth molecular composition of the
        populations from which these reads were sampled from physically.
        '''
        if self.Phase_Set_max is None:
            return
        elif self.Phase_Set_max.phase == 'Nonphasable':
            return 'Nonphasable'
        else:
            return sum([self.bayes_numerator_for_haplotype_i(i) for i in range(1, self.ploidy_phase_set + 1)])




from typing import Dict

hg19: Dict[str, str] = {}
hg38: Dict[str, str] = {}

hg38[
    'NA12877'] = 'https://ftp-trace.ncbi.nlm.nih.gov/giab/ftp/technical/platinum_genomes/2016-1.0/hg38/small_variants/NA12877/NA12877.vcf.gz'
hg38[
    'NA12878'] = 'https://ftp-trace.ncbi.nlm.nih.gov/giab/ftp/technical/platinum_genomes/2016-1.0/hg38/small_variants/NA12878/NA12878.vcf.gz'

hg19[
    'NA12877'] = 'https://ftp-trace.ncbi.nlm.nih.gov/giab/ftp/technical/platinum_genomes/2016-1.0/hg19/small_variants/NA12877/NA12877.vcf.gz'
hg19[
    'NA12878'] = 'https://ftp-trace.ncbi.nlm.nih.gov/giab/ftp/technical/platinum_genomes/2016-1.0/hg19/small_variants/NA12878/NA12878.vcf.gz'

hg38[
    'NA12891'] = 'https://ftp-trace.ncbi.nlm.nih.gov/giab/ftp/technical/10Xgenomics_ChromiumGenome_LongRanger2.1_09302016/NA12891_GRCh38/NA12891_GRCh38_phased_variants.vcf.gz'
hg38[
    'NA12892'] = 'https://ftp-trace.ncbi.nlm.nih.gov/giab/ftp/technical/10Xgenomics_ChromiumGenome_LongRanger2.1_09302016/NA12892_GRCh38/NA12892_GRCh38_phased_variants.vcf.gz'

hg19[
    'NA12891'] = 'https://ftp-trace.ncbi.nlm.nih.gov/giab/ftp/technical/10Xgenomics_ChromiumGenome_LongRanger2.1_09302016/NA12891_hg19/NA12891_hg19_phased_variants.vcf.gz'
hg19[
    'NA12892'] = 'https://ftp-trace.ncbi.nlm.nih.gov/giab/ftp/technical/10Xgenomics_ChromiumGenome_LongRanger2.1_09302016/NA12892_hg19/NA12892_hg19_phased_variants.vcf.gz'

hg38['reference_sequence_file_name'] = 'hg38.fa.gz'
hg19['reference_sequence_file_name'] = 'hg19.fa.gz'
hg38['reference_sequence'] = 'https://hgdownload.soe.ucsc.edu/goldenPath/hg38/bigZips/latest/hg38.fa.gz'
hg19['reference_sequence'] = 'https://hgdownload.soe.ucsc.edu/goldenPath/hg19/bigZips/latest/hg19.fa.gz'

hg38[
    'HG001'] = 'https://ftp-trace.ncbi.nlm.nih.gov/giab/ftp/release/NA12878_HG001/latest/GRCh38/HG001_GRCh38_GIAB_highconf_CG-IllFB-IllGATKHC-Ion-10X-SOLID_CHROM1-X_v.3.3.2_highconf_PGandRTGphasetransfer.vcf.gz'
hg38[
    'HG002'] = 'https://ftp-trace.ncbi.nlm.nih.gov/giab/ftp/release/AshkenazimTrio/HG002_NA24385_son/latest/GRCh38/HG002_GRCh38_1_22_v4.2.1_benchmark.vcf.gz'
hg38[
    'HG003'] = 'https://ftp-trace.ncbi.nlm.nih.gov/giab/ftp/release/AshkenazimTrio/HG003_NA24149_father/latest/GRCh38/HG003_GRCh38_1_22_v4.2.1_benchmark.vcf.gz'
hg38[
    'HG004'] = 'https://ftp-trace.ncbi.nlm.nih.gov/giab/ftp/release/AshkenazimTrio/HG004_NA24143_mother/latest/GRCh38/HG004_GRCh38_1_22_v4.2.1_benchmark.vcf.gz'
hg38[
    'HG005'] = 'https://ftp-trace.ncbi.nlm.nih.gov/giab/ftp/release/ChineseTrio/HG005_NA24631_son/latest/GRCh38/HG005_GRCh38_GIAB_highconf_CG-Illfb-IllsentieonHC-Ion-10XsentieonHC-SOLIDgatkHC_CHROM1-22_v.3.3.2_highconf.vcf.gz'
hg38[
    'HG006'] = 'https://ftp-trace.ncbi.nlm.nih.gov/giab/ftp/release/ChineseTrio/HG006_NA24694_father/latest/GRCh38/HG006_GIAB_GRCh38_highconf_CG-IllFB-IllSNT-10X_CHROM1-22_v.3.3.2_highconf.vcf.gz'
hg38[
    'HG007'] = 'https://ftp-trace.ncbi.nlm.nih.gov/giab/ftp/release/ChineseTrio/HG007_NA24695_mother/latest/GRCh38/HG007_GIAB_GRCh38_highconf_CG-IllFB-IllSNT-10X_CHROM1-22_v.3.3.2_highconf.vcf.gz'

hg19[
    'HG001'] = 'https://ftp-trace.ncbi.nlm.nih.gov/giab/ftp/release/NA12878_HG001/latest/GRCh37/HG001_GRCh37_GIAB_highconf_CG-IllFB-IllGATKHC-Ion-10X-SOLID_CHROM1-X_v.3.3.2_highconf_PGandRTGphasetransfer.vcf.gz'
hg19[
    'HG002'] = 'https://ftp-trace.ncbi.nlm.nih.gov/giab/ftp/release/AshkenazimTrio/HG002_NA24385_son/latest/GRCh37/HG002_GRCh37_1_22_v4.2.1_benchmark.vcf.gz'
hg19[
    'HG003'] = 'https://ftp-trace.ncbi.nlm.nih.gov/giab/ftp/release/AshkenazimTrio/HG003_NA24149_father/latest/GRCh37/HG003_GRCh37_1_22_v4.2.1_benchmark.vcf.gz'
hg19[
    'HG004'] = 'https://ftp-trace.ncbi.nlm.nih.gov/giab/ftp/release/AshkenazimTrio/HG004_NA24143_mother/latest/GRCh37/HG004_GRCh37_1_22_v4.2.1_benchmark.vcf.gz'
hg19[
    'HG005'] = 'https://ftp-trace.ncbi.nlm.nih.gov/giab/ftp/release/ChineseTrio/HG005_NA24631_son/latest/GRCh37/HG005_GRCh37_highconf_CG-IllFB-IllGATKHC-Ion-SOLID_CHROM1-22_v.3.3.2_highconf.vcf.gz'
hg19[
    'HG006'] = 'https://ftp-trace.ncbi.nlm.nih.gov/giab/ftp/release/ChineseTrio/HG006_NA24694_father/latest/GRCh37/HG006_GIAB_GRCh37_highconf_CG-IllFB-IllSNT-10X_CHROM1-22_v.3.3.2_highconf.vcf.gz'
hg19[
    'HG007'] = 'https://ftp-trace.ncbi.nlm.nih.gov/giab/ftp/release/ChineseTrio/HG007_NA24695_mother/latest/GRCh37/HG007_GIAB_GRCh37_highconf_CG-IllFB-IllSNT-10X_CHROM1-22_v.3.3.2_highconf.vcf.gz'
